#define _GNU_SOURCE
#include <arpa/inet.h>
#include <errno.h>
#include <error.h>
#include <fcntl.h>
#include <ifaddrs.h>
#include <inttypes.h>
#include <linux/errqueue.h>
#include <linux/ethtool.h>
#include <linux/net_tstamp.h>
#include <linux/sockios.h>
#include <net/if.h>
#include <netdb.h>
#include <netinet/in.h>
#include <pthread.h>
#include <sched.h>
#include <signal.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#include "send_packet.h"
#include "utilities.h"

static void fill_histograms(packet_info_t *packet_info, int64_t histograms[NB_HISTOGRAMS][MAX_HIST_VAL]);

static char rx_buffer[MAX_BUFFER_SIZE];
static int sock_fd;

static int so_timestamping_flags =
    SOF_TIMESTAMPING_RX_SOFTWARE | SOF_TIMESTAMPING_SOFTWARE;

// Sets the interface
static int set_if(char *network_if) {
  struct ifreq ifreq;

  memset(&ifreq, 0, sizeof(ifreq));
  strncpy(ifreq.ifr_name, network_if, sizeof(ifreq.ifr_name) - 1);

  if (ioctl(sock_fd, SIOCGIFINDEX, &ifreq))
    error(EXIT_FAILURE, errno, "ioctl SIOCGIFINDEX failed\n");

  return ifreq.ifr_ifindex;
}

int init_udp_recv(int use_timestamps, char *network_if) {
  int getaddrinfo_err;
  int set_if_err;
  struct addrinfo hints, *servinfo, *servinfo_it;

  memset(&hints, 0, sizeof hints);
  hints.ai_family = AF_UNSPEC;
  hints.ai_socktype = SOCK_DGRAM;
  hints.ai_flags = AI_PASSIVE;

  getaddrinfo_err = getaddrinfo(NULL, SERVER_PORT, &hints, &servinfo);
  if (getaddrinfo_err != 0) {
    fprintf(stderr, "getaddrinfo: %s\n", gai_strerror(getaddrinfo_err));
    exit(EXIT_FAILURE);
  }

  for (servinfo_it = servinfo; servinfo_it;
       servinfo_it = servinfo_it->ai_next) {
    sock_fd = socket(servinfo->ai_family, servinfo->ai_socktype,
                     servinfo->ai_protocol);

    if (bind(sock_fd, servinfo_it->ai_addr, servinfo_it->ai_addrlen) == -1) {
      close(sock_fd);
      continue;
    }
    break;
  }

  freeaddrinfo(servinfo);

  if (sock_fd == -1)
    error(EXIT_FAILURE, errno, "Couldn't create receive socket");

  set_if_err = set_if(network_if);
  if (set_if_err < 0)
    error(EXIT_FAILURE, errno, "Couldn't set interface\n");

  if (setsockopt(sock_fd, SOL_SOCKET, SO_BINDTODEVICE, network_if,
                 strlen(network_if)))
    error(EXIT_FAILURE, errno, "setsockopt SO_BINDTODEVICE failed\n");

  if (use_timestamps) {
    if (setsockopt(sock_fd, SOL_SOCKET, SO_TIMESTAMPING, &so_timestamping_flags,
                   sizeof(so_timestamping_flags)))
      error(EXIT_FAILURE, errno, "setsockopt SO_TIMESTAMPING failed\n");
  }

  return sock_fd;
}

/*
 * Receives udp packets
 */
packet_info_t recv_udp_packet(int use_timestamps, int use_histograms, int64_t histograms[NB_HISTOGRAMS][MAX_HIST_VAL]) {

  struct cmsghdr *cmsg;
  struct msghdr msg; // Message hardware, sent to the socket
  struct iovec iov;  // The iovec structures stores the RX buffer
  struct sockaddr_in sin;
  struct {
    struct cmsghdr cm;
    char control[512];
  } control;

  int recvmsgerr;

  packet_info_t packet_info;
  struct timespec ts;

  iov.iov_base = &rx_buffer;
  iov.iov_len = MAX_BUFFER_SIZE - 1;

  memset(&msg, 0, sizeof(msg));
  msg.msg_name = &sin;
  msg.msg_namelen = sizeof(sin);
  msg.msg_iov = &iov;
  msg.msg_iovlen = 1;
  msg.msg_control = &control;
  msg.msg_controllen = sizeof(control);

  if (use_timestamps) {
    clock_gettime(CLOCK_REALTIME, &ts);
    packet_info.userspace_exit_ts = ts_to_uint(ts);
  }

  recvmsgerr = recvmsg(sock_fd, &msg, 0);
  if (recvmsgerr < 0)
    error(EXIT_FAILURE, errno, "recvmsg failed, ret value: %d\n", recvmsgerr);

  if (use_timestamps) {
    clock_gettime(CLOCK_REALTIME, &ts);
    packet_info.userspace_enter_ts = ts_to_uint(ts);
  }

  if (use_timestamps) {
    for (cmsg = CMSG_FIRSTHDR(&msg); cmsg; cmsg = CMSG_NXTHDR(&msg, cmsg)) {
      if (cmsg->cmsg_level == SOL_SOCKET && cmsg->cmsg_type == SO_TIMESTAMPING) {

        struct timespec *stamp = (struct timespec *)CMSG_DATA(cmsg);
        packet_info.kernelspace_ts = ts_to_uint(*stamp);

        clock_gettime(CLOCK_REALTIME, &ts);
        packet_info.userspace_exit_ts = ts_to_uint(ts);

        if (use_histograms)
          fill_histograms(&packet_info, histograms);
      }
    }
  }

  strcpy(packet_info.data, rx_buffer);

  return packet_info;
}

static void fill_histograms(packet_info_t *packet_info, int64_t histograms[NB_HISTOGRAMS][MAX_HIST_VAL]) {

  uint64_t user_space_time = packet_info->userspace_exit_ts - packet_info->userspace_enter_ts;
  uint64_t kernel_space_time = packet_info->userspace_enter_ts - packet_info->kernelspace_ts;

  user_space_time /= 1000u;
  kernel_space_time /= 1000u;

  if (user_space_time > MAX_HIST_VAL)
    fprintf(stderr, "user_space_time value too high: %" PRIu64 "us\n", user_space_time);
  else
    histograms[0][user_space_time]++;

  if (kernel_space_time > MAX_HIST_VAL)
    fprintf(stderr, "kernel_space_time value too high: %" PRIu64 "us\n", kernel_space_time);
  else
    histograms[1][kernel_space_time]++;
}

#ifdef DEBUG

/*
 * Code from scheduled_tx_tools
 */
static int process_socket_error_queue() {
  uint8_t msg_control[CMSG_SPACE(sizeof(struct sock_extended_err))];
  unsigned char err_buffer[256];
  struct sock_extended_err *serr;
  struct cmsghdr *cmsg;
  __u64 tstamp = 0;

  struct iovec iov = {.iov_base = err_buffer, .iov_len = sizeof(err_buffer)};
  struct msghdr msg = {.msg_iov = &iov,
                       .msg_iovlen = 1,
                       .msg_control = msg_control,
                       .msg_controllen = sizeof(msg_control)};

  if (recvmsg(sock_fd, &msg, MSG_ERRQUEUE) == -1) {
    fprintf(stderr, "recvmsg failed");
    return -1;
  }

  cmsg = CMSG_FIRSTHDR(&msg);
  while (cmsg != NULL) {
    serr = (void *)CMSG_DATA(cmsg);
    if (serr->ee_origin == SO_EE_ORIGIN_TXTIME) {
      tstamp = ((__u64)serr->ee_data << 32) + serr->ee_info;

      switch (serr->ee_code) {
      case SO_EE_CODE_TXTIME_INVALID_PARAM:
        fprintf(stderr,
                "packet with tstamp %llu dropped due to invalid params\n",
                tstamp);
        return 0;
      case SO_EE_CODE_TXTIME_MISSED:
        fprintf(stderr,
                "packet with tstamp %llu dropped due to missed deadline\n",
                tstamp);
        return 0;
      default:
        return -1;
      }
    }

    cmsg = CMSG_NXTHDR(&msg, cmsg);
  }

  return 0;
}
#endif
